#!/usr/bin/env python

from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
from requests.auth import HTTPBasicAuth
import requests

from a2p2.client import A2P2ClientPreferences
__all__ = []

import logging

logger = logging.getLogger(__name__)

# Do not use next code directly - this is dev related code.
# Please have a look on the other jmmc modules for client code.


#  Retry Handling for requests
# See https://www.peterbe.com/plog/best-practice-with-retries-with-requests
# Or https://findwork.dev/blog/advanced-usage-python-requests-timeouts-retries-hooks/


# TODO check that we return more infiormation for 500 error case because some server side code part do return 500
# and here we may

def requests_retry_session(
        retries=3,
        backoff_factor=1.0,
        status_forcelist=(500, 502, 504)):

    adapter = HTTPAdapter()
    adapter.max_retries = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
        respect_retry_after_header=False
    )
    session = requests.Session()
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session

#            r = requests.request(method, url, headers=headers, data=json.dumps(data))
# -> r = requests_session.request(method, url, headers=headers, data=json.dumps(data))


class JmmcAPI():
    def __init__(self, rootURL, username=None, password=None):
        self.rootURL = rootURL

        # credentials can be given :
        # by parameters
        # if username or password is None, try to get it from the a2p2 preferences ( run: a2p2 -c )
        # if username or password still is None keep auth to None.
        # a None auth will search for ~/.netrc files trying to fix 401 accesses

        prefs = A2P2ClientPreferences()
        if not username :
            username = prefs.getJmmcLogin()
        if not password :
            password = prefs.getJmmcPassword()

        # if we got complete credential
        if username and password:
            self.auth = HTTPBasicAuth(username, password)
        else:
            self.auth = None

        self.requests_session = requests_retry_session()

        logger.debug("Instanciating JmmcAPI on '%s'" % self.rootURL)

    def _get(self, url):
        return self._request('GET', url)

    def _put(self, url, json):
        return self._request('PUT', url, json)

    def _post(self, url, json):
        return self._request('POST', url, json)

    def _post(self, url, **kwargs):
        return self._request2('POST', url, **kwargs)

    def _request2(self, method, url, **kwargs):
        logger.info("performing %s request on %s" % (method, self.rootURL+url))
        r = self.requests_session.request(
            method, self.rootURL+url, **kwargs)
        # handle response if any or throw an exception
        if (r.status_code == 204):  # No Content : everything is fine
            return
        elif 200 <= r.status_code < 300:
            if 'Content-Type' in r.headers.keys() and 'application/json' in r.headers['Content-Type']:
                return r.json()
            else:
                return r.content
        # TODO enhance error handling ? Throw an exception ....
        error = []
        error.append("status_code is %s"%r.status_code)
        if r.reason :
            error.append(r.reason)
        if "X-Http-Error-Description" in r.headers.keys():
            error.append(r.headers["X-Http-Error-Description"])

        raise Exception(error)

    def _request(self, method, url, json=None, data=None, files=None):
        logger.info("performing %s request on %s" % (method, self.rootURL+url))
        r = self.requests_session.request(
            method, self.rootURL+url, auth=self.auth, json=json, data=data, files=files)
        # handle response if any or throw an exception
        if (r.status_code == 204):  # No Content : everything is fine
            return
        elif 200 <= r.status_code < 300:
            if 'Content-Type' in r.headers.keys() and 'application/json' in r.headers['Content-Type']:
                return r.json()
            else:
                return r.content
        # TODO enhance error handling ? Throw an exception ....
        error = []
        error.append("status_code is %s"%r.status_code)
        if r.reason :
            error.append(r.reason)
        if "X-Http-Error-Description" in r.headers.keys():
            error.append(r.headers["X-Http-Error-Description"])

        raise Exception(error)



#!/usr/bin/env python

__all__ = []

import cgi
import re
import traceback
import xml.etree.ElementTree

import numpy as np
from astropy.coordinates import SkyCoord


#
# This code is dead code but still leave the time we validate the new vlti directory code.
#


def parseXmlMessage(client, url, p2container):
    ui = client.ui
    api = client.facilityManager.getAPI()

    ui.setProgress(0)

    currentInstrument = p2container.instrument
    containerId = p2container.containerId

    e = xml.etree.ElementTree.parse(url)

    try:
        interferometerConfiguration = e.find('interferometerConfiguration')
        interferometer = interferometerConfiguration.find('name').text

        if interferometer != "VLTI":
            ui.ShowErrorMessage(
                "ASPRO did not sent obs for VLTI, no action taken.")
            return

        BASELINE = interferometerConfiguration.find('stations').text

        instrumentConfiguration = e.find('instrumentConfiguration')
        instrument = instrumentConfiguration.find('name').text
        if instrument != currentInstrument:
            ui.ShowErrorMessage("ASPRO not set for currently selected instrument: %s , expected %s" %
                                (currentInstrument, instrument))
            return
        # FIXME: TBD CHANGE TO HAVE OTHER INSTRUMENTS THAN GRAVITY!
        supportedInstruments = client.facilityManager.getSupportedInstruments()
        if not instrument in supportedInstruments:
            ui.ShowErrorMessage(
                "ASPRO not set for supported instruments [%s], no action taken." % (', '.join(supportedInstruments)))
            return

        instrumentMode = instrumentConfiguration.find('instrumentMode').text

        # if we have more than 1 obs, then better put it in a subfolder waiting
        # for the existence of a block sequence not yet implemented in P2
        obsconflist = e.findall('observationConfiguration')
        doFolder = (len(obsconflist) > 1)
        parentContainerId = containerId
        if doFolder:
            folderName = (obsconflist[0].find('SCTarget')).find('name').text
            folderName = re.sub('[^A-Za-z0-9]+', '_', folderName.strip())
            folder, _ = api.createFolder(containerId, folderName)
            containerId = folder['containerId']

        for observationConfiguration in e.findall('observationConfiguration'):
            # science. If calibrator, get also DIAMETER (and compute VIS?)

            TYPE = observationConfiguration.find('type').text
            if TYPE == 'SCIENCE':
                OBJTYPE = 'SCIENCE'
            else:
                OBJTYPE = 'CALIBRATOR'

            scienceTarget = observationConfiguration.find('SCTarget')
            NAME = scienceTarget.find('name').text
            SCRA = scienceTarget.find('RA').text
            # RA MUST have 3 digits precision and DEC at least 2!!!
            w = SCRA.rfind('.')
            l = len(SCRA)
            if l - w < 4:
                ui.ShowErrorMessage(
                    "Object " + NAME + " has a too low precision in RA to be useable by VLTI, please correct.")
                return

            if l - w > 4:
                SCRA = SCRA[0:w + 4]
            SCDEC = scienceTarget.find('DEC').text
            w = SCDEC.rfind('.')
            l = len(SCDEC)
            if l - w < 3:
                ui.ShowErrorMessage(
                    "Object " + NAME + " has a too low precision in DEC to be useable by VLTI, please correct.")
                return
            if l - w > 4:
                SCDEC = SCDEC[0:w + 4]

            PMRA = 0.0
            PMDEC = 0.0
            # PMRA and DEC may be null without problem.
            pmratxt = scienceTarget.find('PMRA')
            if pmratxt != None:
                PMRA = float(pmratxt.text)
            pmdetxt = scienceTarget.find('PMDEC')
            if pmdetxt != None:
                PMDEC = float(pmdetxt.text)

            # but these should be defined.
            COU_GS_MAG = float(scienceTarget.find('FLUX_V').text)
            SEQ_INS_SOBJ_MAG = float(scienceTarget.find('FLUX_K').text)
            SEQ_FI_HMAG = float(scienceTarget.find('FLUX_H').text)
            # setup some default values, to be changed below
            COU_AG_GSSOURCE = 'SCIENCE'  # by default
            GSRA = '00:00:00.000'
            GSDEC = '00:00:00.000'
            COU_AG_PMA = 0.0
            COU_AG_PMD = 0.0
            dualField = False
            DIAMETER = 0.0
            VIS = 1.0

            if OBJTYPE == 'CALIBRATOR' and scienceTarget.find('DIAMETER') != None:
                DIAMETER = float(scienceTarget.find('DIAMETER').text)
                VIS = 1.0  # FIXME

            # initialize FT variables (must exist)
            FTRA = ""
            FTDEC = ""
            SEQ_FT_ROBJ_NAME = ""
            SEQ_FT_ROBJ_MAG = -99.99
            SEQ_FT_ROBJ_DIAMETER = -1.0
            SEQ_FT_ROBJ_VIS = -1.0

            # if FT Target is not ScTarget, we are in dual-field (TBD)
            ftTarget = observationConfiguration.find('FTTarget')
            if ftTarget != None:
                try:
                    SEQ_FT_ROBJ_NAME = ftTarget.find('name').text
                    FTRA = ftTarget.find('RA').text
                    w = FTRA.rfind('.')
                    l = len(FTRA)
                    if l - w < 4:
                        ui.ShowErrorMessage(
                            "Object " + SEQ_FT_ROBJ_NAME + " has a too low precision in RA to be useable by VLTI, please correct.")
                        return
                    if l - w > 4:
                        FTRA = FTRA[0:w + 4]
                    FTDEC = ftTarget.find('DEC').text
                    w = FTDEC.rfind('.')
                    l = len(FTDEC)
                    if l - w < 3:
                        ui.ShowErrorMessage(
                            "Object " + SEQ_FT_ROBJ_NAME + " has a too low precision in DEC to be useable by VLTI, please correct.")
                    return
                    if l - w > 4:
                        FTDEC = FTDEC[0:w + 4]
                    # no PMRA, PMDE for FT !!
                    SEQ_FI_HMAG = float(ftTarget.find('FLUX_H').text)
                    # just to say we must treat the case
                    # there is no FT Target
                    SEQ_FT_ROBJ_MAG = SEQ_FI_HMAG
                    SEQ_FT_ROBJ_DIAMETER = 0.0  # FIXME
                    SEQ_FT_ROBJ_VIS = 1.0  # FIXME
                    dualField = True
                except:
                    # print ("incomplete Fringe Tracker Target definition!")
                    ui.ShowErrorMessage(
                        "incomplete Fringe Tracker Target definition, OB not set!")

            # AO target
            aoTarget = observationConfiguration.find('AOTarget')
            if aoTarget != None:
                try:
                    AONAME = aoTarget.find('name').text
                    COU_AG_GSSOURCE = 'SETUPFILE'  # since we have an AO
                    AORA = aoTarget.find('RA').text
                    w = AORA.rfind('.')
                    l = len(AORA)
                    if l - w > 4:
                        AORA = AORA[0:w + 4]
                    AODEC = aoTarget.find('DEC').text
                    w = AODEC.rfind('.')
                    l = len(AODEC)
                    if l - w > 4:
                        AODEC = AODEC[0:w + 4]

                    COU_AG_PMA = 0.0
                    COU_AG_PMD = 0.0
                    # PMRA and DEC may be null without problem.
                    pmratxt = aoTarget.find('PMRA')
                    if pmratxt != None:
                        COU_AG_PMA = float(pmratxt.text)
                    pmdetxt = aoTarget.find('PMDEC')
                    if pmdetxt != None:
                        COU_AG_PMD = float(pmdetxt.text)

                except:
                    # print ("incomplete Adaptive Optics Target definition!")
                    ui.ShowErrorMessage(
                        "incomplete Adaptive Optics Target definition, OB not set!")

            # Guide Star
            gsTarget = observationConfiguration.find('GSTarget')
            if gsTarget != None:
                try:
                    GSNAME = gsTarget.find('name').text
                    COU_AG_SOURCE = 'SETUPFILE'  # since we have an GS
                    GSRA = gsTarget.find('RA').text
                    w = GSRA.rfind('.')
                    l = len(GSRA)
                    if l - w > 4:
                        GSRA = GSRA[0:w + 4]
                    GSDEC = gsTarget.find('DEC').text
                    w = GSDEC.rfind('.')
                    l = len(GSDEC)
                    if l - w > 4:
                        GSDEC = GSDEC[0:w + 4]
                    COU_GS_MAG = float(gsTarget.find('FLUX_V').text)
                    # no PMRA, PMDE for GS !!

                except:
                    # print ("incomplete GuideStar Target definition!")
                    ui.ShowErrorMessage(
                        "incomplete GuideStar Target definition, OB not set!")

            # LST interval
            try:
                obsConstraint = observationConfiguration.find(
                    'observationConstraints')
                LSTINTERVAL = obsConstraint.find('LSTinterval').text
            except:
                LSTINTERVAL = None

            # then call the ob-creation using the API.
            createGravityOB(
                ui, client.getUsername(
                ), api, containerId, OBJTYPE, NAME, BASELINE, instrumentMode, SCRA, SCDEC, PMRA, PMDEC,
                SEQ_INS_SOBJ_MAG, SEQ_FI_HMAG, DIAMETER,
                COU_AG_GSSOURCE, GSRA, GSDEC, COU_GS_MAG, COU_AG_PMA, COU_AG_PMD, dualField, FTRA, FTDEC,
                SEQ_FT_ROBJ_NAME, SEQ_FT_ROBJ_MAG, SEQ_FT_ROBJ_DIAMETER, SEQ_FT_ROBJ_VIS, LSTINTERVAL)
            ui.addToLog("Processed: " + NAME)
        # endfor
        if doFolder:
            containerId = parentContainerId
            doFolder = False
    except Exception as e:
        trace = traceback.format_exc(limit=1)
        ui.ShowErrorMessage(
            "General error or Absent Parameter in template!\n Missing magnitude or OB not set ?\n\nError :\n %s " % (
                trace))
        ui.setProgress(0)


# here dit must be a string since this is what p2 expects. NOT an integer
# or real/double.


def getDit(mag, spec, pol, tel, mode):
    string_dit = "1"

    if mode == 1:  # Dual
        if tel == 1:
            mag -= 3.7  # UT, DUAL
        else:
            mag -= 0.7  # AT, DUAL
    elif tel == 1:
        mag -= 3.0  # UT, SINGLE

    if spec == 2:  # HR
        if pol == 1:  # SPLIT
            if mag > 1:
                string_dit = "30"
            elif mag > 0:
                string_dit = "10"
            elif mag > -0.5:
                string_dit = "5"
            else:
                string_dit = "3"
        else:  # COMB
            if mag > 2:
                string_dit = "30"
            elif mag > 0.5:
                string_dit = "10"
            elif mag > -0.5:
                string_dit = "5"
            else:
                string_dit = "1"
    elif spec == 1:  # MR
        if pol == 1:  # SPLIT
            if mag > 4:
                string_dit = "30"
            elif mag > 3:
                string_dit = "10"
            elif mag > 2.5:
                string_dit = "5"
            elif mag > 1.5:
                string_dit = "3"
            elif mag > 0.0:
                string_dit = "1"
            else:
                string_dit = "0.3"
        else:  # COMB
            if mag > 5:
                string_dit = "30"
            elif mag > 3.5:
                string_dit = "10"
            elif mag > 3.0:
                string_dit = "5"
            elif mag > 2.5:
                string_dit = "3"
            elif mag > 1.0:
                string_dit = "1"
            else:
                string_dit = "0.3"
    elif spec == 0:  # LR FIXME VALUES ARE NOT GIVEN!!!!!!!!
        if pol == 1:  # SPLIT
            if mag > 9:
                string_dit = "30"
            elif mag > 7:
                string_dit = "10"
            elif mag > 6.5:
                string_dit = "5"
            elif mag > 5.5:
                string_dit = "3"
            elif mag > 4.0:
                string_dit = "1"
            else:
                string_dit = "0.3"
        else:  # COMB
            if mag > 10:
                string_dit = "30"
            elif mag > 9:
                string_dit = "10"
            elif mag > 7.5:
                string_dit = "5"
            elif mag > 6.5:
                string_dit = "3"
            elif mag > 5.0:
                string_dit = "1"
            else:
                string_dit = "0.3"
    return string_dit


# define function creating the OB:


def createGravityOB(ui, username, api, containerId, OBJTYPE, NAME, BASELINE, instrumentMode, SCRA, SCDEC, PMRA, PMDEC,
                    SEQ_INS_SOBJ_MAG, SEQ_FI_HMAG, DIAMETER, COU_AG_GSSOURCE, GSRA, GSDEC, COU_GS_MAG, COU_AG_PMA,
                    COU_AG_PMD, dualField, FTRA, FTDEC, SEQ_FT_ROBJ_NAME, SEQ_FT_ROBJ_MAG, SEQ_FT_ROBJ_DIAMETER,
                    SEQ_FT_ROBJ_VIS, LSTINTERVAL):
    ui.setProgress(0.1)
    # UT or AT?
    isUT = (BASELINE[0] == "U")
    if isUT:
        SCtoREFmaxDist = 2000
        SCtoREFminDist = 400
        tel = 1
        if SEQ_INS_SOBJ_MAG < 5:
            skyTransparencyConstrainText = 'Variable, thin cirrus'
        else:
            skyTransparencyConstrainText = 'Clear'
    else:
        SCtoREFminDist = 1500
        SCtoREFmaxDist = 4000
        tel = 0
        if SEQ_INS_SOBJ_MAG < 3:
            skyTransparencyConstrainText = 'Variable, thin cirrus'
        else:
            skyTransparencyConstrainText = 'Clear'

    VISIBILITY = 1.0
    dualmode = 0
    # if Dualfield and offset between fields is bad, complain and do Nothing
    diff = [0, 0]
    if dualField:
        dualmode = 1
        # compute x,y between science and ref beams:
        diff = getSkyDiff(SCRA, SCDEC, FTRA, FTDEC)
        if np.abs(diff[0]) < SCtoREFminDist:
            ui.ShowErrorMessage("Dual-Field distance of two stars is  < " + str(
                SCtoREFminDist) + " mas, Please Correct.")
            return
        elif np.abs(diff[0]) > SCtoREFmaxDist:
            ui.ShowErrorMessage("Dual-Field distance of two stars is  > " + str(
                SCtoREFmaxDist) + " mas, Please Correct.")
            return

    if instrumentMode == 'LOW-COMBINED':
        INS_SPEC_RES = 'LOW'
        INS_FT_POL = 'OUT'
        INS_SPEC_POL = 'OUT'
        string_dit = getDit(SEQ_INS_SOBJ_MAG, 0, 0, tel, dualmode)
    elif instrumentMode == 'LOW-SPLIT':
        INS_SPEC_RES = 'LOW'
        INS_FT_POL = 'IN'
        INS_SPEC_POL = 'IN'
        string_dit = getDit(SEQ_INS_SOBJ_MAG, 0, 1, tel, dualmode)
    elif instrumentMode == 'MEDIUM-COMBINED':
        INS_SPEC_RES = 'MED'
        INS_FT_POL = 'OUT'
        INS_SPEC_POL = 'OUT'
        string_dit = getDit(SEQ_INS_SOBJ_MAG, 1, 0, tel, dualmode)
    elif instrumentMode == 'MEDIUM-SPLIT':
        INS_SPEC_RES = 'MED'
        INS_FT_POL = 'IN'
        INS_SPEC_POL = 'IN'
        string_dit = getDit(SEQ_INS_SOBJ_MAG, 1, 1, tel, dualmode)
    elif instrumentMode == 'HIGH-COMBINED':
        INS_SPEC_RES = 'HIGH'
        INS_FT_POL = 'OUT'
        INS_SPEC_POL = 'OUT'
        string_dit = getDit(SEQ_INS_SOBJ_MAG, 2, 0, tel, dualmode)
    elif instrumentMode == 'HIGH-SPLIT':
        INS_SPEC_RES = 'HIGH'
        INS_FT_POL = 'IN'
        INS_SPEC_POL = 'IN'
        string_dit = getDit(SEQ_INS_SOBJ_MAG, 2, 1, tel, dualmode)
    else:
        ui.ShowErrorMessage("Invalid Instrument Mode, Please Correct.")
        return

    # compute ndit, nexp
    dit = float(string_dit)
    ndit = 300 / dit
    if ndit < 10:
        ndit = 10
    if ndit > 300:
        ndit = 300
    ndit = int(ndit)  # must be integer
    exptime = int(ndit * dit + 40)  # 40 sec overhead by exp
    nexp = (1800 - 900) / exptime
    nexp = int(nexp)
    ui.addToLog('number of exposures to reach 1800 s per OB is ' + str(nexp))
    if nexp < 3:
        nexp = 3  # min is O S O
        # recompute ndit
        exptime = (1800 - 900) / nexp
        ndit = (exptime - 40) / dit
        ndit = int(ndit)
        if ndit < 10:
            ndit = 10
            ui.addToLog(
                "**Warning**, OB NDIT has been set to min value=10, but OB will take longer than 1800 s")
    nexp %= 40
    sequence = 'O S O O S O O S O O S O O S O O S O O S O O S O O S O O S O O S O O S O O S O O'
    my_sequence = sequence[0:2 * nexp]

    # everything seems OK
    # create new OB in container:
    goodName = re.sub('[^A-Za-z0-9]+', '_', NAME.strip())
    if OBJTYPE == 'SCIENCE':
        isCalib = False
        OBS_DESCR = 'SCI_' + goodName + '_GRAVITY_' + \
                    BASELINE.replace(' ', '') + '_' + instrumentMode
    else:
        isCalib = True
        OBS_DESCR = 'CAL_' + goodName + '_GRAVITY_' + \
                    BASELINE.replace(' ', '') + '_' + instrumentMode
    ob, obVersion = api.createOB(containerId, OBS_DESCR)
    obId = ob['obId']

    # we use obId to populate OB
    ob['obsDescription']['name'] = OBS_DESCR[0:min(len(OBS_DESCR), 31)]
    ob['obsDescription']['userComments'] = 'Generated by ' + \
                                           username + ' using ASPRO 2 (c) JMMC'
    # ob['obsDescription']['InstrumentComments'] = 'AO-B1-C2-E3' #should be a
    # list of alternative quadruplets!

    ob['target']['name'] = NAME.replace(' ', '_')
    ob['target']['ra'] = SCRA
    ob['target']['dec'] = SCDEC
    ob['target']['properMotionRa'] = round(PMRA / 1000.0, 4)
    ob['target']['properMotionDec'] = round(PMDEC / 1000.0, 4)

    ob['constraints']['name'] = 'Aspro-created constraint'
    # FIXME: error (OB): "Phase 2 constraints must closely follow what was requested in the Phase 1 proposal.
    # The seeing value allowed for this OB is >= java0x0 arcsec."
    ob['constraints']['seeing'] = 1.0
    ob['constraints']['skyTransparency'] = skyTransparencyConstrainText
    ob['constraints']['baseline'] = BASELINE.replace(' ', '-')
    # FIXME: default values NOT IN ASPRO!
    # ob['constraints']['airmass'] = 5.0
    # ob['constraints']['fli'] = 1

    ob, obVersion = api.saveOB(ob, obVersion)

    # LST constraints if present
    # by default, above 40 degree. Will generate a WAIVERABLE ERROR if not.
    if LSTINTERVAL:
        sidTCs, stcVersion = api.getSiderealTimeConstraints(obId)
        lsts = LSTINTERVAL.split('/')
        lstStartSex = lsts[0]
        lstEndSex = lsts[1]
        # p2 seems happy with endlst < startlst
        # a = SkyCoord(lstStartSex+' +0:0:0',unit=(u.hourangle,u.deg))
        # b = SkyCoord(lstEndSex+' +0:0:0',unit=(u.hourangle,u.deg))
        # if b.ra.deg < a.ra.deg:
        # api.saveSiderealTimeConstraints(obId,[ {'from': lstStartSex, 'to': '00:00'},{'from': '00:00','to': lstEndSex}], stcVersion)
        # else:
        api.saveSiderealTimeConstraints(
            obId, [{'from': lstStartSex, 'to': lstEndSex}], stcVersion)

    ui.setProgress(0.2)

    # then, attach acquisition template(s)
    if dualField:
        tpl, tplVersion = api.createTemplate(obId, 'GRAVITY_dual_acq')
        # put values
        tpl, tplVersion = api.setTemplateParams(obId, tpl, {
            'SEQ.FT.ROBJ.NAME': SEQ_FT_ROBJ_NAME,
            'SEQ.FT.ROBJ.MAG': round(SEQ_FT_ROBJ_MAG, 3),
            'SEQ.FT.ROBJ.DIAMETER': SEQ_FT_ROBJ_DIAMETER,
            'SEQ.FT.ROBJ.VIS': SEQ_FT_ROBJ_VIS,
            'SEQ.FT.MODE': "AUTO",
            'SEQ.INS.SOBJ.NAME': NAME,
            'SEQ.INS.SOBJ.MAG': round(SEQ_INS_SOBJ_MAG, 3),
            'SEQ.INS.SOBJ.DIAMETER': DIAMETER,
            'SEQ.INS.SOBJ.VIS': VISIBILITY,
            'SEQ.INS.SOBJ.X': diff[0],
            'SEQ.INS.SOBJ.Y': diff[1],
            'SEQ.FI.HMAG': round(SEQ_FI_HMAG, 3),
            'TEL.TARG.PARALLAX': 0.0,
            'INS.SPEC.RES': INS_SPEC_RES,
            'INS.FT.POL': INS_FT_POL,
            'INS.SPEC.POL': INS_SPEC_POL,
            'COU.AG.GSSOURCE': COU_AG_GSSOURCE,
            'COU.AG.ALPHA': GSRA,
            'COU.AG.DELTA': GSDEC,
            'COU.GS.MAG': round(COU_GS_MAG, 3),
            'COU.AG.PMA': round(COU_AG_PMA / 1000, 4),
            'COU.AG.PMD': round(COU_AG_PMD / 1000, 4)
        }, tplVersion)

    else:
        tpl, tplVersion = api.createTemplate(obId, 'GRAVITY_single_acq')
        # put values
        tpl, tplVersion = api.setTemplateParams(obId, tpl, {
            'SEQ.INS.SOBJ.NAME': NAME,
            'SEQ.INS.SOBJ.MAG': round(SEQ_INS_SOBJ_MAG, 3),
            'SEQ.INS.SOBJ.DIAMETER': DIAMETER,
            'SEQ.INS.SOBJ.VIS': VISIBILITY,
            'COU.AG.GSSOURCE': COU_AG_GSSOURCE,
            'COU.AG.ALPHA': GSRA,
            'COU.AG.DELTA': GSDEC,
            'COU.GS.MAG': round(COU_GS_MAG, 3),
            'COU.AG.PMA': round(COU_AG_PMA / 1000, 4),
            'COU.AG.PMD': round(COU_AG_PMD / 1000, 4),
            'SEQ.FI.HMAG': round(SEQ_FI_HMAG, 3),
            'TEL.TARG.PARALLAX': 0.0,
            'INS.SPEC.RES': INS_SPEC_RES,
            'INS.FT.POL': INS_FT_POL,
            'INS.SPEC.POL': INS_SPEC_POL
        }, tplVersion)

    templateId = tpl['templateId']

    ui.setProgress(0.3)

    if isCalib:
        if dualField:
            tpl, tplVersion = api.createTemplate(
                obId, 'GRAVITY_dual_obs_calibrator')
        else:
            tpl, tplVersion = api.createTemplate(
                obId, 'GRAVITY_single_obs_calibrator')
    else:
        if dualField:
            tpl, tplVersion = api.createTemplate(obId, 'GRAVITY_dual_obs_exp')
        else:
            tpl, tplVersion = api.createTemplate(
                obId, 'GRAVITY_single_obs_exp')
    templateId = tpl['templateId']

    ui.setProgress(0.4)

    # put values. they are the same except for dual obs science (?)
    if dualField and not isCalib:
        tpl, tplVersion = api.setTemplateParams(obId, tpl, {
            'DET2.DIT': string_dit,
            'DET2.NDIT.OBJECT': ndit,
            'DET2.NDIT.SKY': ndit,
            'SEQ.OBSSEQ': my_sequence,
            'SEQ.SKY.X': 2000,
            'SEQ.SKY.Y': 2000
        }, tplVersion)
    else:
        tpl, tplVersion = api.setTemplateParams(obId, tpl, {
            'DET2.DIT': string_dit,
            'DET2.NDIT.OBJECT': ndit,
            'DET2.NDIT.SKY': ndit,
            'SEQ.OBSSEQ': my_sequence,
            'SEQ.RELOFF.X': "0.0",
            'SEQ.RELOFF.Y': "0.0",
            'SEQ.SKY.X': 2000,
            'SEQ.SKY.Y': 2000
        }, tplVersion)

    ui.setProgress(0.5)

    # verify OB online
    response, _ = api.verifyOB(obId, True)

    ui.setProgress(1.0)

    if response['observable']:
        ui.ShowInfoMessage('OB ' + str(obId) + ' ' + ob['name'] + ' is OK.')
        ui.addToLog('OB: ' + str(obId) + ' is ok')
    else:
        s = ""
        for ss in response['messages']:
            s += cgi.escape(ss) + '\n'
        ui.ShowWarningMessage(
            'OB ' + str(obId) + ' <b>HAS Warnings</b>. ESO says:\n\n' + s)
        ui.addToLog('OB: ' + str(obId) + ' created with warnings')
        # (NOTE: we need to escape things like <= in returned text)

        # fetch OB again to confirm its status change
        #   ob, obVersion = api.getOB(obId)
        # python3: print('Status of verified OB', obId, 'is now',
        # ob['obStatus'])

    def getSkyDiff(ra, dec, ftra, ftdec):
        science = SkyCoord(ra, dec, frame='icrs', unit='deg')
        ft = SkyCoord(ftra, ftdec, frame='icrs', unit='deg')
        ra_offset = (science.ra - ft.ra) * np.cos(ft.dec.to('radian'))
        dec_offset = (science.dec - ft.dec)
        return [ra_offset.deg * 3600 * 1000, dec_offset.deg * 3600 * 1000]  # in mas



from .models import _model

def punct( name, flux_weight=1.0, x=0.0, y=0.0,  output_mode=None):
    """ Returns the Fourier transform of a punctual object (Dirac function) at coordinates (X,Y)
(milliarcsecond).

FLUX_WEIGHT is the intensity coefficient. FLUX_WEIGHT=1 means total energy is 1. """

    return _model({"flux_weight" : flux_weight, "x" : x, "y" : y, "name":name, "type":"punct"}, output_mode)

def disk( name, flux_weight=1.0, x=0.0, y=0.0, diameter=0.0,  output_mode=None):
    """ Returns the Fourier transform of a normalized uniform disk of diameter DIAMETER
(milliarcsecond) and centered at coordinates (X,Y) (milliarcsecond).

FLUX_WEIGHT is the intensity coefficient. FLUX_WEIGHT=1 means total energy is 1.

The function returns an error if DIAMETER is negative. """

    return _model({"flux_weight" : flux_weight, "x" : x, "y" : y, "diameter" : diameter, "name":name, "type":"disk"}, output_mode)

def elong_disk( name, flux_weight=1.0, x=0.0, y=0.0, minor_axis_diameter=0.0, elong_ratio=1.0, major_axis_pos_angle=0.0,  output_mode=None):
    """ Returns the Fourier transform of a normalized ellipse centered at coordinates (X,Y)
(milliarcsecond) with a ratio ELONG_RATIO between the major diameter and the minor one
MINOR_AXIS_DIAMETER, turned from the positive vertical semi-axis (i.e. North direction)
with angle MAJOR_AXIS_POS_ANGLE, in degrees, towards to the positive horizontal semi-axis
(i.e. East direction). (the elongation is along the major_axis)

For avoiding degenerescence, the domain of variation of MAJOR_AXIS_POS_ANGLE is 180
degrees, for ex. from 0 to 180 degrees.

ELONG_RATIO = major_axis / minor_axis
FLUX_WEIGHT is the intensity coefficient. FLUX_WEIGHT=1 means total energy is 1.

The function returns an error if MINOR_AXIS_DIAMETER is negative or if ELONG_RATIO is
smaller than 1. """

    return _model({"flux_weight" : flux_weight, "x" : x, "y" : y, "minor_axis_diameter" : minor_axis_diameter, "elong_ratio" : elong_ratio, "major_axis_pos_angle" : major_axis_pos_angle, "name":name, "type":"elong_disk"}, output_mode)

def flatten_disk( name, flux_weight=1.0, x=0.0, y=0.0, major_axis_diameter=0.0, flatten_ratio=1.0, minor_axis_pos_angle=0.0,  output_mode=None):
    """ Returns the Fourier transform of a normalized ellipse centered at coordinates (X,Y)
(milliarcsecond) with a ratio FLATTEN_RATIO between the major diameter
MAJOR_AXIS_DIAMETER and the minor one, turned from the positive vertical semi-axis
(i.e. North direction) with angle MINOR_AXIS_POS_ANGLE, in degrees, towards to the
positive horizontal semi-axis (i.e. East direction). (the flattening is along the minor_axis)

For avoiding degenerescence, the domain of variation of MINOR_AXIS_POS_ANGLE is 180
degrees, for ex. from 0 to 180 degrees.

FLATTEN_RATIO = major_axis / minor_axis
FLUX_WEIGHT is the intensity coefficient. FLUX_WEIGHT=1 means total energy is 1.

The function returns an error if MAJOR_AXIS_DIAMETER is negative or if FLATTEN_RATIO
is smaller than 1. """

    return _model({"flux_weight" : flux_weight, "x" : x, "y" : y, "major_axis_diameter" : major_axis_diameter, "flatten_ratio" : flatten_ratio, "minor_axis_pos_angle" : minor_axis_pos_angle, "name":name, "type":"flatten_disk"}, output_mode)

def circle( name, flux_weight=1.0, x=0.0, y=0.0, diameter=0.0,  output_mode=None):
    """ Returns the Fourier transform of a normalized uniform circle of diameter DIAMETER
(milliarcsecond) and centered at coordinates (X,Y) (milliarcsecond).

FLUX_WEIGHT is the intensity coefficient. FLUX_WEIGHT=1 means total energy is 1.

The function returns an error if DIAMETER is negative. """

    return _model({"flux_weight" : flux_weight, "x" : x, "y" : y, "diameter" : diameter, "name":name, "type":"circle"}, output_mode)

def ring( name, flux_weight=1.0, x=0.0, y=0.0, diameter=0.0, width=0.0,  output_mode=None):
    """ Returns the Fourier transform of a normalized uniform ring with internal diameter
DIAMETER (milliarcsecond) and external diameter DIAMETER + WIDTH centered at coordinates
(X,Y) (milliarcsecond).

FLUX_WEIGHT is the intensity coefficient. FLUX_WEIGHT=1 means total energy is 1.

The function returns an error if DIAMETER or WIDTH are negative. """

    return _model({"flux_weight" : flux_weight, "x" : x, "y" : y, "diameter" : diameter, "width" : width, "name":name, "type":"ring"}, output_mode)

def elong_ring( name, flux_weight=1.0, x=0.0, y=0.0, minor_internal_diameter=0.0, elong_ratio=1.0, width=0.0, major_axis_pos_angle=0.0,  output_mode=None):
    """ Returns the Fourier transform of a normalized uniform elongated ring centered at
coordinates (X,Y) (milliarcsecond). The sizes of the function in two orthogonal directions
are given by the narrowest internal diameter (MINOR_INTERNAL_DIAMETER) and by the ratio
ELONG_RATIO between the widest internal diameter and MINOR_INTERNAL_DIAMETER,
in the same way as for an ellipse (the elongation is along the major_axis) :

ELONG_RATIO = MAJOR_INTERNAL_DIAMETER / MINOR_INTERNAL_DIAMETER.
In the direction of MINOR_INTERNAL_DIAMETER, the external diameter is
MINOR_INTERNAL_DIAMETER + WIDTH. In the direction of the widest internal diameter,
the width is magnified by the ratio ELONG_RATIO, so that the external diameter is
the elongated MAJOR_INTERNAL_DIAMETER + WIDTH * ELONG_RATIO.
MAJOR_AXIS_POS_ANGLE is measured in degrees, from the positive vertical semi-axis
(i.e. North direction) towards to the positive horizontal semi-axis (i.e. East direction).
For avoiding degenerescence, the domain of variation of MAJOR_AXIS_POS_ANGLE is 180
degrees, for ex. from 0 to 180 degrees.

FLUX_WEIGHT is the intensity coefficient. FLUX_WEIGHT=1 means total energy is 1.

The function returns an error if MINOR_INTERNAL_DIAMETER is negative or if ELONG_RATIO
is smaller than 1. """

    return _model({"flux_weight" : flux_weight, "x" : x, "y" : y, "minor_internal_diameter" : minor_internal_diameter, "elong_ratio" : elong_ratio, "width" : width, "major_axis_pos_angle" : major_axis_pos_angle, "name":name, "type":"elong_ring"}, output_mode)

def flatten_ring( name, flux_weight=1.0, x=0.0, y=0.0, major_internal_diameter=0.0, flatten_ratio=1.0, width=0.0, minor_axis_pos_angle=0.0,  output_mode=None):
    """ Returns the Fourier transform of a normalized uniform flattened ring centered at
coordinates (X,Y) (milliarcsecond). The sizes of the function in two orthogonal directions
are given by the widest internal diameter (MAJOR_INTERNAL_DIAMETER) and by the ratio
FLATTEN_RATIO between MAJOR_INTERNAL_DIAMETER and the narrowest internal diameter,
in the same way as for an ellipse (the flattening is along the minor axis) :

FLATTEN_RATIO = MAJOR_INTERNAL_DIAMETER / MINOR_INTERNAL_DIAMETER.
In the direction of MAJOR_INTERNAL_DIAMETER, the external diameter is
MAJOR_INTERNAL_DIAMETER + WIDTH. In the direction of the narrowest internal diameter,
the width is decreased by the ratio FLATTEN_RATIO, so that the external diameter is
the flattened MINOR_INTERNAL_DIAMETER + WIDTH / FLATTEN_RATIO.
MINOR_AXIS_POS_ANGLE is measured in degrees, from the positive vertical semi-axis
(i.e. North direction) towards to the positive horizontal semi-axis (i.e. East direction).
For avoiding degenerescence, the domain of variation of MINOR_AXIS_POS_ANGLE is 180
degrees, for ex. from 0 to 180 degrees.

FLUX_WEIGHT is the intensity coefficient. FLUX_WEIGHT=1 means total energy is 1.

The function returns an error if MAJOR_INTERNAL_DIAMETER is negative or if FLATTEN_RATIO
is smaller than 1. """

    return _model({"flux_weight" : flux_weight, "x" : x, "y" : y, "major_internal_diameter" : major_internal_diameter, "flatten_ratio" : flatten_ratio, "width" : width, "minor_axis_pos_angle" : minor_axis_pos_angle, "name":name, "type":"flatten_ring"}, output_mode)

def gaussian( name, flux_weight=1.0, x=0.0, y=0.0, fwhm=0.0,  output_mode=None):
    """ Returns the Fourier transform of a normalized gaussian with given FWHM (milliarcsecond)
centered at coordinates (X,Y) (milliarcsecond).

FLUX_WEIGHT is the intensity coefficient. FLUX_WEIGHT=1 means total energy is 1.

The function returns an error if FWHM is negative. """

    return _model({"flux_weight" : flux_weight, "x" : x, "y" : y, "fwhm" : fwhm, "name":name, "type":"gaussian"}, output_mode)

def elong_gaussian( name, flux_weight=1.0, x=0.0, y=0.0, minor_axis_fwhm=0.0, elong_ratio=1.0, major_axis_pos_angle=0.0,  output_mode=None):
    """ Returns the Fourier transform of a normalized elongated gaussian centered at coordinates
(X,Y) (milliarcsecond). The sizes of the function in two orthogonal directions are given by
the narrowest FWHM (MINOR_AXIS_FWHM) and by the ratio ELONG_RATIO between the largest
FWHM (MAJOR_AXIS_FWHM) and the MINOR_AXIS_FWHM, in the same way as for an ellipse
(the elongation is along the major_axis) :

ELONG_RATIO = MAJOR_AXIS_FWHM / MINOR_AXIS_FWHM.
MAJOR_AXIS_POS_ANGLE is measured in degrees, from the positive vertical semi-axis
(i.e. North direction) towards to the positive horizontal semi-axis (i.e. East direction).
For avoiding degenerescence, the domain of variation of MAJOR_AXIS_POS_ANGLE is 180
degrees, for ex. from 0 to 180 degrees.

FLUX_WEIGHT is the intensity coefficient. FLUX_WEIGHT=1 means total energy is 1.

The function returns an error if MINOR_AXIS_FWHM is negative or if ELONG_RATIO
is smaller than 1. """

    return _model({"flux_weight" : flux_weight, "x" : x, "y" : y, "minor_axis_fwhm" : minor_axis_fwhm, "elong_ratio" : elong_ratio, "major_axis_pos_angle" : major_axis_pos_angle, "name":name, "type":"elong_gaussian"}, output_mode)

def flatten_gaussian( name, flux_weight=1.0, x=0.0, y=0.0, major_axis_fwhm=0.0, flatten_ratio=1.0, minor_axis_pos_angle=0.0,  output_mode=None):
    """ Returns the Fourier transform of a normalized flattened gaussian centered at coordinates
(X,Y) (milliarcsecond). The sizes of the function in two orthogonal directions are given by
the largest FWHM (MAJOR_AXIS_FWHM) and by the ratio FLATTEN_RATIO between the largest
FWHM (MAJOR_AXIS_FWHM) and the MINOR_AXIS_FWHM, in the same way as for an ellipse
(the flattening is along the minor_axis) :

FLATTEN_RATIO = MAJOR_AXIS_FWHM / MINOR_AXIS_FWHM.
MINOR_AXIS_POS_ANGLE is measured in degrees, from the positive vertical semi-axis
(i.e. North direction) towards to the positive horizontal semi-axis (i.e. East direction).
For avoiding degenerescence, the domain of variation of MINOR_AXIS_POS_ANGLE is 180
degrees, for ex. from 0 to 180 degrees.

FLUX_WEIGHT is the intensity coefficient. FLUX_WEIGHT=1 means total energy is 1.

The function returns an error if MAJOR_AXIS_FWHM is negative or if FLATTEN_RATIO
is smaller than 1. """

    return _model({"flux_weight" : flux_weight, "x" : x, "y" : y, "major_axis_fwhm" : major_axis_fwhm, "flatten_ratio" : flatten_ratio, "minor_axis_pos_angle" : minor_axis_pos_angle, "name":name, "type":"flatten_gaussian"}, output_mode)

def limb_quadratic( name, flux_weight=1.0, x=0.0, y=0.0, diameter=0.0, a1_coeff=0.0, a2_coeff=0.0,  output_mode=None):
    """ Returns the Fourier transform of a center-to-limb darkened disk of diameter DIAMETER
(milliarcsecond) centered at coordinates (X,Y) (milliarcsecond).

The brightness distribution o, if expressed versus mu, the cosine of the azimuth of
a surface element of the star, follows a quadratic law of coefficients
A1_COEFF, A2_COEFF ([-1,1]), and is normalized for mu = 1 (center of the star).
o(mu) = 1 -A1_COEFF(1-mu) - A2_COEFF(1-mu)^2.

FLUX_WEIGHT is the intensity coefficient. FLUX_WEIGHT=1 means total energy is 1.

The function returns an error if DIAMETER is negative or if A1_COEFF or A2_coeff is
outside bounds [-1,1] """

    return _model({"flux_weight" : flux_weight, "x" : x, "y" : y, "diameter" : diameter, "a1_coeff" : a1_coeff, "a2_coeff" : a2_coeff, "name":name, "type":"limb_quadratic"}, output_mode)




#!/usr/bin/env python

__all__ = ['A2p2Client']

import time
import traceback
import os
import configparser
import logging

from a2p2 import __version__
from a2p2.facility import FacilityManager
from a2p2.gui import MainWindow
from a2p2.ob import OB
from a2p2.samp import A2p2SampClient
from a2p2.vlti.facility import VltiFacility



# prepare global logging
a2p2Rootlogger = logging.getLogger("a2p2")
a2p2Rootlogger.setLevel(logging.INFO)
# uncomment next two lines to log requests done by p2api as debug and maybe other ones...
#a2p2Rootlogger = logging.getLogger()
#a2p2Rootlogger.setLevel(logging.DEBUG)
console = logging.StreamHandler()
console.setLevel(logging.DEBUG)
consoleFormatter = logging.Formatter(
    '%(levelname)s - %(name)s  - %(asctime)s - %(filename)s:%(lineno)d - %(message)s')
console.setFormatter(consoleFormatter)
a2p2Rootlogger.addHandler(console)

logger = logging.getLogger(__name__)


class A2p2Client():
    """Transmit your Aspro2 observation to remote Observatory scheduling database.

       with A2p2Client() as a2p2:
           a2p2.run()
           ..."""

    def __init__(self, fakeAPI=False, verbose=False):
        """Create the A2p2 client."""

        self.preferences = A2P2ClientPreferences()

        self.apiName = ""
        if fakeAPI:
            self.apiName = "fakeAPI"
        self.fakeAPI = fakeAPI

        if verbose:
            a2p2Rootlogger.setLevel(logging.DEBUG)

        self.ui = MainWindow(self)
        # Instantiate the samp client and connect to the hub later
        self.a2p2SampClient = A2p2SampClient()
        self.facilityManager = FacilityManager(self)

        if self.preferences.exists():
            pass
        else:
            self.ui.addToLog(
                "No preference file found, please create one so your data persists (launch program with -c option).\n")

        self.ui.addToLog("Please often update ( pip install -U [--user] a2p2 ) and don't hesitate to send any feedback or issues!\n\n")


        self.errors = []

        pass

    def __enter__(self):
        """Handle starting the 'with' statements."""

        self.ui.addToLog("              Welcome in the A2P2 V" + __version__)
        self.ui.addToLog("")
        self.ui.addToLog("( https://github.com/JMMC-OpenDev/a2p2/wiki )")

        return self

    def __del__(self):
        """Handle deleting the object."""
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        """Handle closing the 'with' statement."""
        del self.a2p2SampClient
        del self.ui
        # TODO close the connection to the obs database ?

        # WARNING do not return self only. Else exceptions are hidden
        # return self

    def __str__(self):
        instruments = "\n- ".join(["Supported instruments:", "TBD"])
        apis = "\n- ".join(["Supported APIs:", "TBD"])
        return """a2p2 client\n%s\n%s\n""" % (instruments, apis)

    def changeSampStatus(self, connected_flag):
        self.sampConnected = connected_flag

    def setProgress(self, perc, actionName=None):
        if actionName:
            print("%s action progress is  %s" % (actionName, perc))
        else:
            print("progress is  %s %%" % (perc))

    def clearErrors(self):
        self.errors.clear()

    def getErrors(self):
        return self.errors

    def addError(self, error):
        self.errors.append(error)

    def processOB(self, ob):
        self.facilityManager.processOB(ob)

    def run(self):
        # bool of status change
        flag = [0]

        logger.info("Running client ...")

        # handle autologin
        if self.preferences.getP2AutoLoginBoolean():
            logger.debug("Autologin using '%s' file" %
                         A2P2ClientPreferences.getPreferencesFileName())
            self.ui.loop()
            vltifacility = self.facilityManager.facilities.get(
                VltiFacility.getName())
            vltifacility.autologin()

        # We now run the loop to wait for the message in a try/finally block so that if
        # the program is interrupted e.g. by control-C, the client terminates
        # gracefully. ( but sometimes fails inside tkinter/__init__.py CallWrapper/__call__() )

        # We test every 1s to see if the hub has sent a message
        delay = 0.1
        each = 10
        loop_cnt = 0
        warnForAspro = True

        while loop_cnt >= 0:
            try:
                loop_cnt += 1
                time.sleep(delay)

                self.ui.loop()

                if not self.a2p2SampClient.is_connected() and loop_cnt % each == 0:
                    try:
                        self.a2p2SampClient.connect()
                        self.ui.setSampId(self.a2p2SampClient.get_public_id())
                    except:
                        self.ui.setSampId(None)
                        if warnForAspro:
                            warnForAspro = False
                            self.ui.addToLog(
                                "\nPlease launch Aspro2 to submit your OBs.")
                        # TODO test for other exception than SAMPHubError(u'Unable to find a running SAMP Hub.',)
                        pass

                if self.a2p2SampClient.has_message():
                    try:
                        if self.a2p2SampClient.has_ob_message():
                            ob = OB(self.a2p2SampClient.get_ob_url())
                            self.facilityManager.processOB(ob)
                    except:
                        self.ui.addToLog(
                            "Exception during a2p2 ob creation: " + traceback.format_exc(), False)
                        self.ui.addToLog("Can't process last OB")

                    try:
                        if self.a2p2SampClient.has_model_message():
                                m = self.a2p2SampClient.get_model()
                                self.showModel(m)
                    except:
                        self.ui.addToLog(
                            "Exception during a2p2 ob creation: " + traceback.format_exc(), False)
                        self.ui.addToLog("Can't process last OB")

                    # always clear previous received message
                    self.a2p2SampClient.clear_message()

                if self.ui.requestAbort:
                    loop_cnt = -1
            except KeyboardInterrupt:
                loop_cnt = -1
        print("\nDisconnecting SAMP ...")
        self.a2p2SampClient.disconnect()
        print("Bye!")


    def createPreferencesFile():
        A2P2ClientPreferences.createPreferencesFile()


    def showModel(self, xmlmodel):
        from a2p2.jmmc.models import modelsFromXml
        self.ui.addToLog("Received star model")
        self.ui.addToLog(modelsFromXml(xmlmodel))


class A2P2ClientPreferences():
    # define application name
    appname = "a2p2"

    def __init__(self):
        self._config = A2P2ClientPreferences.getPreferences()
        pass

    def exists(self):
        if self._config.sections():
            return True
        else:
            return False

    def getPreferences():
        preferences_file = A2P2ClientPreferences.getPreferencesFileName()
        config = configparser.ConfigParser(allow_no_value=True)
        config.read(preferences_file)
        return config

    def getPreferencesFileName():
        from appdirs import user_config_dir
        preferences_file = os.path.join(user_config_dir(
            A2P2ClientPreferences.appname), "prefs.ini")
        return preferences_file

    def createPreferencesFile():
        filename = A2P2ClientPreferences.getPreferencesFileName()
        if os.path.exists(filename):
            print(f"{filename} already exists. Nothing done")
            p=A2P2ClientPreferences()
            versionInPref=p.getA2P2Version()
            if __version__ != versionInPref:
                print(f"💡 You may try to backup this file (V{versionInPref}) and merge with a new generated one for V{__version__}.")
        else:
            import getpass

            config = configparser.ConfigParser(allow_no_value=True)
            #config['DEFAULT'] = {'_noprefyet': '42'}

            config['a2p2'] = {}
            s = config['a2p2']
            s['# A2P2 SECTION'] = ""
            s['# = > please do not modify next properties unless you handle the possible cAtAstrOphE 🔥 ? <'] = ""
            s['version'] = __version__


            config['jmmc'] = {}
            s = config['jmmc']
            s['# JMMC SECTION'] = ""
            s['# = > please uncomment and update next properties to make it active <'] = ""
            s['#login'] = "my.email@my.lab"
            s['#password'] = "12345zZ"

            config['p2'] = {}
            s=config['p2']
            s['# ESO P2 SECTION'] = ""
            s['# = > please uncomment and update next properties to make it active <'] = ""
            s['#username'] = getpass.getuser()
            s['#password'] = "12345zZ"
            s['#autologin'] = "yes"
            s['# = > please change next properties to change the name recorded in the OB comments <'] = ""
            s['user_comment_name'] = f"{getpass.getuser()}"

            config['p2.iss.vltitype'] = {}
            s=config['p2.iss.vltitype']
            s['# = > please uncomment the default values to add for ISS.VLTITYPE <'] = ""
            s['# = > all supported will be added for any VLTI instrument if info is not provided by Aspro2 <'] = None
            s['#snapshot'] = None
            s['#imaging'] = None
            s['#time-series'] = None
            s['#astrometry'] = None

            os.makedirs(os.path.dirname(filename), exist_ok=True)
            with open(filename, 'w+') as configfile:
                config.write(configfile)

            print("'%s' template file created to store preferences. Please check and adjust it's content." % filename)

    def getConfig(self, section, key, default=None):
        try:
            return self._config[section][key]
        except:
            return default

    def getConfigKeys(self, section, default=None):
        try:
            return list(self._config[section].keys())
        except:
            return default


    def getConfigBoolean(self, section, key, default=None):
        try:
            return self._config.getboolean(section, key)
        except:
            return default

    # Retrieve A2P2 prefs
    def getA2P2Version(self):
        return self.getConfig("a2p2", "version", 'missing')

    # Retrieve JMMC prefs
    def getJmmcLogin(self):
        return self.getConfig("jmmc", "login", None)

    def getJmmcPassword(self):
        return self.getConfig("jmmc", "password", None)


    # Retrieve P2 prefs
    def getP2Username(self):
        """ Get P2 username in config or use default demo account """
        return self.getConfig("p2", "username", '52052')

    def getP2Password(self):
        """Get P2 password in config or use default demo account"""
        return self.getConfig("p2", "password", 'tutorial')

    def getP2UserCommentName(self):
        import getpass
        return self.getConfig("p2", "user_comment_name", getpass.getuser())

    def getP2AutoLoginBoolean(self):
        return self.getConfigBoolean("p2", "autologin", False)





#!/usr/bin/env python

__all__ = []

import logging
import json
import xml.etree.ElementTree as ET
from collections import defaultdict, namedtuple, OrderedDict
import re

import copy


logger = logging.getLogger(__name__)

# https://stackoverflow.com/questions/2148119/how-to-convert-an-xml-string-to-a-dictionary-in-python
# see comment below for our custom mods on attributes naming
def etree_to_dict(t):
    d = {t.tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {t.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
        # print("add d=%s"%str(d))
    if t.attrib:
        attrs = {}
        for k, v in t.attrib.items():
            if not "{" in k[0]: # ignore elementName with prefix
                attrs[k]=v
        d[t.tag].update(attrs)
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[t.tag]['#text'] = text
        else:
            d[t.tag] = text
    return d

def normalizeObj(obj):
    """
    Modify in place the structure if required.
    Note:
        dict key must be valid identifier (only tested for EXTRA_INFORMATIONS created fields)
    """
    otype=type(obj)
    if otype is dict:
        for k,v in obj.items():

            if k in ['observationConfiguration', 'HAinterval', 'LSTinterval'] and not isinstance(v, list):
                # always defines observationConfiguration as a list
                normalizeObj(v)
                obj[k]=[v]
            elif k=='EXTRA_INFORMATIONS':
                print(f"Updating dict {k} was : {obj[k]}")
                # replace fields by dict entries
                try:
                    fields = obj[k]["parameter"]+obj[k]["field"]
                    for vn in fields:
                        fname=vn["name"]
                        # test if we have to normalize field name that will become a python field
                        if fname.isidentifier():
                            obj[k][fname]=vn["value"]
                        else:
                            # replace unvalid chars by _ and append F to avoid starting by _
                            nfname=re.sub('^_','F_', re.sub('\W|^(?=\d)','_', fname))
                            obj[k][nfname]=vn["value"]


                    del obj[k]["field"]
                except:
                    pass
            else:
                normalizeObj(v)
        return obj
    elif otype is list:
        for e in obj:
            normalizeObj(e)
    else:
        pass

class OB():
    """
    Give access to OBs sent by Aspro2 and provide some helper functions.
    You can use attributes to walk to associated configurations values, e.g.:
    - ob.interferometerConfiguration.name
    - ob.instrumentConfiguration
    - ob.observationConfiguration[]
    - ob.observationSchedule

    every values are string (must be converted for numeric values).

    Following fields are converted to list even if the list get a single element. The OB code can then handle such multiples values.
    ['observationConfiguration', 'HAinterval', 'LSTinterval']

    """

    def __init__(self, url):
        # extract XML in elementTree
        e = ET.parse(url)
        # convert data to dict structure
        d = etree_to_dict(e.getroot())
        # keep only content of subelement to avoid schema version change
        # '{http://www.jmmc.fr/aspro-ob/0.1}observingBlockDefinition'
        ds = d[list(d)[0]]  # -> version and name are lost

        #self.srcDict  = copy.deepcopy(ds)
        normalizeObj(ds)
        #self.normDict = ds
        #self.srcObj  = toObject(self.srcDict,inputName="OB")
        #self.normObj  = toObject(self.normDict,inputName="OB")

        # store subelements as attributes
        for e in ds.keys():
            # parse JSON into an object with attributes corresponding to dict keys.
            # We should probably avoid json use...
            o = json.loads(
                json.dumps(ds[e]), object_hook=lambda d: namedtuple(e, d.keys())(*d.values()))
            setattr(self, e, o)

        # store normalized source for str repr
        self.ds = ds

    def getFluxes(self, target):
        """
        Return a flux mesurements as dict (ordered dict by BVRIJHK ).
        """
        order = "BVRIJHK"
        fluxes = {}
        els = target._asdict()
        for k in els.keys():
            if k.startswith("FLUX_"):
                fluxes[k[5:]] = els[k]  # remove FLUX_ prefix for dict key

        return OrderedDict(sorted(fluxes.items(), key=lambda t: order.find(t[0])))

    def getPeriod(self):
        """
        Return period given by Aspro2 as integer.
        """
        # use re to remove alpha characters from "Period 109" and make.
        return int(re.sub(r"\D", "", self.interferometerConfiguration.version))


    def get(self, obj, fieldname, defaultvalue=None):
        if fieldname in obj._fields:
            return getattr(obj, fieldname)
        else:
            return defaultvalue

    def __str__(self):
        if True:
            import json
            return json.dumps(self.ds, indent=2)
        else:
            return str(self.ds)



#!/usr/bin/env python

__all__ = []

import logging

from ..client import A2P2ClientPreferences

from .utils import JmmcAPI
from . import PRODLABEL

logger = logging.getLogger(__name__)


class Catalog():
    """ Get remote access to read and update catalogs exposed through JMMC's API.
        Credential can be explicitly given for method that require an authentication, else:
        - a2p2 preferences login will be used if present (see a2p2 -c)
        - or uses .netrc file
    """

    def __init__(self, catalogName, username=None, password=None, prod=False, apiUrl=None):
        self.catalogName = catalogName
        self.prod = prod

        # Manage prod & preprod or user provided access points
        if apiUrl:
            self.apiUrl = apiUrl  # trust given url as catalogAPI if value is provided
        elif self.prod:
            self.apiUrl = "https://oidb.jmmc.fr/restxq/catalogs"
        else:
            self.apiUrl = "https://oidb-beta.jmmc.fr/restxq/catalogs"

        self.api = JmmcAPI(self.apiUrl, username, password)

        logger.info(f"Create catalog wrapper to access '{catalogName}' ({PRODLABEL[self.prod]} API at {self.api.rootURL})")

    def list(self):
        """ Get list of exposed catalogs on API associated to this catalog. """
        return self.api._get("")

    def metadata(self):
        """ Get catalog metadata """
        return self.api._get(f"/meta/{self.catalogName}")

    def pis(self):
        """ Get PIs from catalog and check for associated JMMC login in OiDB datapi table."""
        return self.api._get(f"/accounts/{self.catalogName}")["pi"]

    def piname(self, jmmcLogin=None):
        """ Get the piname associated to the given jmmcLogin.
            If jmmcLogin parameter is not provided, try to get jmmc.login preferences.
            Only pi names returned by pis() can be retrieved.

            usage: piname()
        """
        if not jmmcLogin:
            prefs = A2P2ClientPreferences()
            jmmcLogin = prefs.getJmmcLogin()
        if not jmmcLogin:
            raise Exception(
                "missing login parameter or jmmc.login preference.")
        pis = self.pis()
        for pi in pis:
            if "login" in pi and pi["login"] == jmmcLogin:
                return pi["name"]

    def getRow(self, id):
        """ Get a single catalog record for the given id.

            usage: cat.getRow(42)
        """
        return self.api._get(f"/{self.catalogName}/{id}")

    def updateRow(self, id, values):
        """ Update record identified by given id and associated values.

            usage: cat.updateRows(42, {"col_a":"a", "col_b":"b" })
        """
        return self.api._put(f"/{self.catalogName}/{id}", values)

    def updateRows(self, values):
        """ Update multiple rows.
            Values must contain a list of dictionnary and each entry must contains id key among other columns.

            usage: updateRows([ { "id":42, "col_a":"a" }, { "id":24, "col_b":"b" } ])
        """

        # We may check befere sending payload that we always provide an id for every record
        return self.api._put(f"/{self.catalogName}", values)

    def addRows(self, values):
        """ Add multiple rows.
            Values is an array of row to add. id column values will be ignored.

            usage: addCatalogRows([ { "id":42, "col_a":"a" }, { "id":24, "col_b":"b" } ])
        """
        return self.api._post(f"/{self.catalogName}", json=values)

    def getDelegations(self, pi=None):
        """ Get -all- delegations.
            A specific pi parameter may be given but requires admin priviledges. Else use current piname will be used by remote service.

            usage: getDelegations()
        """
        res = self.api._get(f"/delegations?pi={pi}")
        if "delegations" in res:
            return res["delegations"] # login and pi values not returned

        return []

    def updateDelegations(self, delegations):
        """ Update delegations according given array of delegations.
            A specific "pi" key may be given in delegations dict but requires admin privileges. Else current piname will be used by remote service.
            "action" may be provided with "add" or "remove" value to set proper delegations (default add given delegation).

            usage: addDelegations([{"pi":"mypiname", "catalogs":"mycats", "coi":["alice","bob"]}])

        """
        return self.api._put("/delegations", json=delegations)

    def removeDelegations(self, delegations):
        """ Remove delegations according given array of delegations.
            A specific pi key may be given but requires admin privileges. Else current piname will be used by remote service.
            usage: removeDelegations([{}, {}])
        """
        return "TODO"





#!/usr/bin/env python

__all__ = []

import sys
import traceback
import logging

from a2p2.gui import FacilityUI

if sys.version_info[0] == 2:
    from Tkinter import *
    from tkMessageBox import *
else:
    from tkinter import *
    from tkinter.messagebox import *

# Constants
_HR = "\n----------------------------------------------\n"

logger = logging.getLogger(__name__)

class CharaUI(FacilityUI):

    def __init__(self, a2p2client):
        FacilityUI.__init__(self, a2p2client)
        # first version store all in a single widget
        self.text = Text(self, width=120)
        scroll = Scrollbar(self, command=self.text.yview)
        self.text.configure(yscrollcommand=scroll.set)
        scroll.pack(side=RIGHT, fill=Y)
        self.text.pack(side=LEFT, fill=BOTH, expand=True)
        # more control could be added in the futur in this area for CHARA
        # specific

        # avoid repeat of baseline on successive schedules
        self.lastBaselines = "-"

    def get(self, obj, fieldname):
        if fieldname in obj._fields:
            return getattr(obj, fieldname)
        else:
            return None

    def displayOB(self, ob):
        try:
            buffer = self.extractReport(ob)
        except:
            buffer = "Error during report generation\n" + \
                     traceback.format_exc() + _HR + str(ob)

        self.text.insert(END, buffer)

    def extractReport(self, ob):
        """ We coud try to mimic the output below
----------------------------------------------
Baselines: S2(2)-E2(4)
           Ref Cart: BL 2
Times in [brackets] are when target is above 30 deg and has delay
----------------------------------------------
S2(2)E2(4)
----------------------------------------------
Start-5:00 UT [Start-7:58 UT]
Object:
HD 78209 (A3V, 30pc): V=4.45, R=4.21, tht=0.54
Fringe Finder:
HD 79158: V=5.29, R=5.29, tht=0.19 (good for fringe finding)
AO Flat Star:
HD 82328: V=3.18
Cals:
1) HD 77309: V=5.73, R=5.65, tht=0.22
2) HD 79158: V=5.29, R=5.29, tht=0.19 (good for fringe finding)
----------------------------------------------
5:00-6:30 UT [Start-9:44 UT]
Object:
HD 91312 (A7IV, IRx, 35pc): V=4.72, R=3.76, tht=0.58
AO Flat Star:
"""
        buffer = ""

        # Display baselines on change
        stations = ob.interferometerConfiguration.stations
        if self.lastBaselines != stations:
            buffer += _HR
            buffer += "Baselines: " + stations + "\n"
            self.lastBaselines = stations
        buffer += _HR

        # Retrieve all stars (as obsConf) and build sciences list
        sciences = []
        targets = {}  # store  ids for futur retrieval in schedule
        for oc in ob.observationConfiguration:
            targets[oc.id] = oc
            if "SCI" in oc.type:
                sciences.append(oc)

        # Retrieve cals from schedule
        cals = {}
        for schedule in ob.observationSchedule.OB:
            try:  # hack for single element observationSchedule
                ref = schedule.ref
            except:
                ref = schedule
            target = targets[ref]
            if "CAL" in target.type:
                cals[ref] = target

        # TODO check for calibrator only ?

        for oc in sciences:
            sct = oc.SCTarget
            extrainfos = self.get(sct, "EXTRA_INFORMATIONS")
            ftt = self.get(oc, "FTTarget")
            aot = self.get(oc, "AOTarget")

            buffer += oc.observationConstraints.LSTinterval + "\n"
            buffer += "Object:\n"
            fluxes = ", ".join([e[0] + "=" + e[1]
                                for e in ob.getFluxes(sct).items()])
            try:
                info = sct.SPECTYP
            except:
                info = "no SPECTYP"
            info+=", "
            try:
                info += sct.PARALLAX
            except:
                info += "no PARALLAX"


            buffer += sct.name + " (" + info + ") : " + fluxes + "\n"
            if ftt:
                buffer += "Fringe Finder:\n"
                fluxes = ", ".join([e[0] + "=" + e[1]
                                    for e in ob.getFluxes(ftt).items()])
                buffer += ftt.name + " : " + fluxes + "\n"
            if aot:
                buffer += "AO Flat Star:\n"
                fluxes = ", ".join([e[0] + "=" + e[1]
                                    for e in ob.getFluxes(aot).items()])
                buffer += aot.name + " : " + fluxes + "\n"

            if len(cals) >= 1:
                buffer += "Cals:\n"
                for cal in cals:
                    buffer += "- " + cal + "\n"

            # Display Extra_Informations if any
            if extrainfos:
                buffer += "Extra_infos:\n"
                for field in extrainfos._fields:
                    buffer += "- " + field + "=" + getattr(extrainfos,field) +"\n"

            buffer += _HR

        return buffer
